/*
    FEARLUS model0-6-2: LandUse.m
    Copyright (C) 1999-2002  Macaulay Institute
    Copyright (C) 1999-2002  Macaulay Institute
    Copyright (C) 1999-2002  Macaulay Institute

    This file is part of FEARLUS model0-6-2, an agent-based model of land use
    change.

    FEARLUS model0-6-2 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-6-2 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the LandUse class.

*/

#import "LandUse.h"
#import "BitString.h"
#import "LandUseHistory.h"
#import "LandParcel.h"
#import "FearlusArguments.h"
#import "Parameter.h"
#import "Environment.h"


@implementation LandUse

/*

create

The superclass method is over-ridden to create a unique identifier for each
member of the class. This is achieved using a static local variable in the
create method. Static variables in methods are shared across the whole class.

*/

+create: (id)z {
  LandUse *new;
  static unsigned counter = 1;
  char buf[1000];

  new = [super create: z];
  new->pin = counter;
  sprintf(buf, "%u", new->pin);
  new->label = (char *)malloc(strlen(buf) + 4);
  if(new->label == NULL) {
    fprintf(stderr, "Memory allocation error\n");
    exit(1);
  }
  sprintf(new->label, "lu%s", buf);
  counter++;
  return(new);
}

/*

getPIN

This is an accessor method for the Personal Identification Number of the
instance of the class. 

*/

-(unsigned)getPIN {
  return(pin);
}

/*

setParameters:

Pass in the model parameters to the object

*/

-(void)setParameters: (Parameter *)p {
  parameter = p;
}

/*

initialiseWithBitLength:colour:environment:

This method should be called after the land use has been created (using
SwarmObject's create method -- we're not doing anything special here). It takes
the length of the land use bit string as an argument, and creates the match and
don't care bit strings. It should only be called once. It also creates the
usageHistory list, and the match and don't care bit strings. The setParameters
method must be called before this one.

*/

-(void)initialiseWithColour: (int)col
		environment: (Environment *)env;
{
  match = [BitString create: [self getZone]
		     setBitCount: [parameter useBitStringSize]];
  parcelMatch = [BitString create: [self getZone]
			   setBitCount: [parameter parcelBitStringSize]];
  dontCare = [BitString create: [self getZone]
			setBitCount: [parameter useBitStringSize]];
  parcelDontCare = [BitString create: [self getZone]
			      setBitCount: [parameter parcelBitStringSize]];
  usageHistory = [List create: [self getZone]];
				// The above will cause memory leaks
				// if this method is called more than
				// once.
  colour = col;
  nLandParcels = 0;
  environment = env;
}

/*

getColour

Return the colour this land use should be printed as

*/

-(int)getColour {
  return colour;
}

/*

getLUMatchWith

This method is used to compute the degree of match between a bit string (which
must be of the same length as the land use bit strings). The degree of match
is calculated in the following way:

LandUse	(match)		1 0 ... 1	1 1 ... 0	0 1 ... 0
  Parity
LandParcel		1 1 ... 0
Climate					1 0 ... 1
Economy							1 1 ... 0
  			-----------------------------------------
  =			1 0 ... 0	1 0 ... 0	0 1 ... 1
  OR
LandUse (dontcare)	0 0 ... 1	0 0 ... 1	0 1 ... 0
  			-----------------------------------------
  =			1 0 ... 1	1 0 ... 1	0 1 ... 1
  Number of 1s		1 1 ... 2	3 3 ... 4	4 5 ... 6

Count the number of 1s in the final result (6 above), and that is the degree
of match.

N.B. This method must be called with a BitString that is a concatenation of
the land parcel, climate and economy -- IN THAT ORDER.

And, yes, this does have to have a silly name because otherwise the compiler
gets confused with BitString's getMatchWith.

*/

-(unsigned)getLUMatchWith: (BitString *)aBitString {
  BitString *result1, *result2;
  id tempZone;
  unsigned degree;
  
  tempZone = [Zone create: scratchZone];
  result1 = [BitString create: tempZone setBitCount: [match getBitCount]];
  result2 = [BitString create: tempZone setBitCount: [match getBitCount]];
  [result1 setToParityOf: match and: aBitString];
  [result2 setToOROf: result1 and: dontCare];
  degree = [result2 countOnes];
  if([FearlusArguments debugInnerCoreMethods]) {
    printf("U: %s: Breakdown of match calculation:\n", sel_get_name(_cmd));
    printf("\t(a)\tParcel+Climate+Economy  : ");
    [aBitString print];
    printf("\t(b)\tLand Use match bits     : ");
    [match print];
    printf("\t(c)\tParity of (a) and (b)   : ");
    [result1 print];
    printf("\t(d)\tLand Use don't care bits: ");
    [dontCare print];
    printf("\t(e)\tOR of (c) and (d)       : ");
    [result2 print];
    printf("\tMatch\tNumber of 1s in (e)     : %u\n", degree);
  }
  [tempZone drop];
  return(degree);
}

/*

getParcelMatchWith:

Match just the parcel bits of the land use with the land parcel given as
argument (i.e. no Climate/Economy data is taken into account)

*/

-(unsigned)getParcelMatchWith: (LandParcel *)lp {
  BitString *result1, *result2;
  id tempZone;
  unsigned degree;

  tempZone = [Zone create: scratchZone];
  result1 = [BitString create: tempZone
		       setBitCount: [parameter parcelBitStringSize]];
  result2 = [BitString create: tempZone
		       setBitCount: [parameter parcelBitStringSize]];
  [result1 setToParityOf: parcelMatch and: [lp getAttributes]];
  [result2 setToOROf: result1 and: parcelDontCare];
  degree = [result2 countOnes];
  if([FearlusArguments debugInnerCoreMethods]) {
    printf("U: %s: Land use %u has match bit string ", sel_get_name(_cmd),
	   pin);
    [match printToFile: stdout];
    printf(" and don't care bit string ");
    [dontCare print];
    printf("\t(a)\tParcel Attributes      : ");
    [[lp getAttributes] print];
    printf("\t(b)\tParcel Match in LU     : ");
    [parcelMatch print];
    printf("\t(c)\tParity of (a) and (b)  : ");
    [result1 print];
    printf("\t(d)\tParcel Don't Care in LU: ");
    [parcelDontCare print];
    printf("\t(e)\tOR of (c) and (d)      : ");
    [result2 print];
    printf("\tMatch\tNumber of 1s in (e)    : %u\n", degree);
  }
  [tempZone drop];
  return(degree);
}

/*

countClimateEconomyDontCare1s

Return a count of the number of 1s in the portion of the Don't Care bit string
that is matched against the climate and economy.

*/

-(unsigned)countClimateEconomyDontCare1s {
  return ( [dontCare countOnes] - [parcelDontCare countOnes] );
}

/*

getMatchWithLandParcel

As per getLUMatchWith, but this time the land parcel, economy and climate are
passed in separately. Construct a bitString consisting of the land parcel,
climate and economy, and call the getLUMatchWith method to get the match.

*/

-(unsigned)
     getMatchWithLandParcel: (LandParcel *)aLandParcel
		    climate: (BitString *)aClimate
		    economy: (BitString *)anEconomy
{
  BitString *climateEconomy, *everything;
  id tempZone;
  unsigned degree;
  
  tempZone = [Zone create: scratchZone];
  climateEconomy = [BitString create: tempZone
			      setBitCount: ([aClimate getBitCount] 
					    + [anEconomy getBitCount])];
  everything = [BitString create: tempZone setBitCount: [match getBitCount]];
  [climateEconomy setToConcatenationOf: aClimate and: anEconomy];
  [everything setToConcatenationOf: [aLandParcel getAttributes]
	      and: climateEconomy];
  degree = [self getLUMatchWith: everything];
  if([FearlusArguments debugUses]) {
    printf("U: Match of land use %u with parcel %u (at %d, %d) = %u (LP=",
	   pin, [aLandParcel getPIN], [aLandParcel getX], [aLandParcel getY],
	   degree);
    [[aLandParcel getAttributes] printToFile: stdout];
    printf(", C=");
    [aClimate printToFile: stdout];
    printf(", E=");
    [anEconomy printToFile: stdout];
    printf(", LP+C+E=");
    [everything printToFile: stdout];
    printf(", UM=");
    [match printToFile: stdout];
    printf(", UDC=");
    [dontCare print];
  }
  [tempZone drop];
  return(degree);
}

/*

randomiseBitsWithDontCare

Set the bit strings to random values. The match bit string is set to have
equal probabilities for 1 and 0. The argument is the probability that the
dontCare bit string has 1.

*/

-(void)randomiseBitsWithDontCare: (double)aProbability {
  [match setEachBitTo: 1 withProbability: 0.5];
  [dontCare setEachBitTo: 1 withProbability: aProbability];
}

/*

updateParcelPortions

Any time the bit strings of a land use are set, the updateParcelPortions method
must be called to update the bit strings that store just the parcel portions of
the match and don't care bit strings. These are used in the getParcelMatchWith:
method.

*/

-(void)updateParcelPortions {
  if([parameter parcelBitStringSize] > 0) {
    [parcelMatch setToSubBitsOf: match
		 from: 0
		 to: [parameter parcelBitStringSize] - 1];
    [parcelDontCare setToSubBitsOf: dontCare
		    from: 0
		    to: [parameter parcelBitStringSize] - 1];
  }
}

/*

recordHistory

Record the history of the land use for the current year

*/

-(void)recordHistory {
  LandUseHistory *hd;

  hd = [LandUseHistory create: [self getZone]];
  [hd initialiseWithYear: [environment getYear] counter: nLandParcels];
  [usageHistory addFirst: hd];
  if([parameter dumpHistory]
     && [parameter maxHistoryLength] < [usageHistory getCount]) {
    [[usageHistory removeLast] drop];
  }
  rememberNLandParcels = nLandParcels;
  // Remember the number of land parcels, so it can be got from the observer
  nLandParcels = 0;
}

/*

writeLandUseHistoryToFile

Print a report of the land use history to the specified file. The history list
is traversed in reverse order because that is the way it is built up -- see the
comments to the addMe:ToLandUseHistoryForYear: method.

*/

-(void)writeLandUseHistoryToFile: (FILE *)fp {
  id indx;
  LandUseHistory *hd;

  for(indx = [usageHistory begin: scratchZone], [indx setLoc: End],
	[indx prev];
      [indx getLoc] == Member;
      [indx prev]) {
    fprintf(fp, "Use\t%u\t", pin);
    hd = [indx get];
    [hd printToFile: fp];
    fprintf(fp, "\t'");
    [match printToFile: fp];
    fprintf(fp, "\t'");
    [dontCare printToFile: fp];
    fprintf(fp, "%s", [FearlusArguments nl]);
  }
  [indx drop];
}

/*

getNLandParcels

Return the number of land parcels for this year that use this land use. This
method is for the observer swarm really -- it is designed to return the number
of land parcels as calculated at the end of the cycle. So, in fact, if called
during a model cycle (rather than display) you will get the LAST number of
land parcels.

*/

-(unsigned)getNLandParcels {
  return(rememberNLandParcels);
}

/*

incNLandParcels

Increment the number of land parcels using this land use

*/

-(void)incNLandParcels {
  nLandParcels++;
}

/*

drop

Override the superclass method so that all items created by an instance of this
class are also dropped.

*/

-(void)drop {
  free(label);
  [match drop];
  [dontCare drop];
  [usageHistory deleteAll];
  [super drop];
}

/*

getLabel

Return a string containing "luPIN", where PIN is replaced with the PIN number
of the land use.

*/

-(char *)getLabel {
  return label;
}

/*

getMatchBitString

Method used by environment if in debug mode to print the match string of the
land uses as it creates them. Probably shouldn't be used anywhere else.

*/

-(BitString *)getMatchBitString {
  return match;
}

/*

getDontCareBitString

Method used by environment if in debug mode to print the don't care string of
the land uses as it creates them. Probably shouldn't be used anywhere else.

*/

-(BitString *)getDontCareBitString {
  return dontCare;
}

/*

getScore

Return the score of the land use. The score is used during numbers of strategy
selection methods (note -- this means that only one land manager can make a
decision at a time -- not a problem, since we're not doing this in parallel
yet).

*/

-(double)getScore {
  if(getScoreReturnsAverage) {
    return score / (double)nScores;
  }
  else {
    return score;
  }
}

/*

initialiseScore:

Method for initialising the scores. This is called from the SelectUseBucket
create method.

*/

-(void)initialiseScore {
  nScores = 0;
  getScoreReturnsAverage = NO;
  score = 0.0;
}

/*

setScore:

Set the score to a particular value. 

*/

-(void)setScore: (double)s {
  score = s;
}

/*

addToScore:

Add a value to the score.

*/

-(void)addToScore: (double)a {
  getScoreReturnsAverage = NO;
  if([FearlusArguments debugInnerCoreMethods]) {
    printf("U: Score of Land Use %u incremented by %g from %g to %g\n",
	   pin, a, score, score + a);
  }
  score += a;
}

/*

addToScoreAverage:

Add a value to the score. getScore will return an average over the scores
entered so far (since the last call to setScore:)

*/

-(void)addToScoreAverage: (double)a {
  getScoreReturnsAverage = YES;
  if([FearlusArguments debugInnerCoreMethods]) {
    printf("U: Score of Land Use %u incremented by %g. Average was %g / %d ="
	   " %g and is now %g / %d = %g\n", pin, a, score, nScores,
	   score / (double)nScores, (score + a), (nScores + 1), 
	   (score + a) / (double)(nScores + 1));
  }
  nScores++;
  score += a;
}


@end
