/*
    FEARLUS model0-6-2: LandUseReport.m
    Copyright (C) 1999-2002  Macaulay Institute
    Copyright (C) 1999-2002  Macaulay Institute
    Copyright (C) 1999-2002  Macaulay Institute

    This file is part of FEARLUS model0-6-2, an agent-based model of land use
    change.

    FEARLUS model0-6-2 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-6-2 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


The implementation for the LandUseReport object

*/

#import "LandUseReport.h"
#import "LandUse.h"
#import "Parameter.h"
#import "ModelSwarm.h"
#import "FearlusArguments.h"
#import "Number.h"
#import "Environment.h"
#import "LandParcel.h"
#import "AssocArray.h"

@implementation LandUseReport

/*

create:

Creation method. Create the land use counter using an
AssocArray.



*/

+(id <Report>)create: aZone {
  LandUseReport *r;

  r = [super create: aZone];
  r->landUseCtr = [AssocArray createBegin: aZone];
  [r->landUseCtr setSize: 7];
  r->landUseCtr = [r->landUseCtr createEnd];

  return r;
}

/*

setModelSwarm:andParameters:

Set the model swarm and parameters for this report. These will be used
to obtain the data in the report. Initialise the land use counter
associative array. It needs a number object to store an integer to be
associated with each land use. The land uses will be the keys of the
associative array.

*/

-(void)setModelSwarm: (ModelSwarm *)m andParameters: (Parameter *)p {
  int i;
  id lu;

  parameter = p;
  model = m;

  lu = [[m getEnvironment] getLandUses];
  for(i = 0; i < [parameter nUses]; i++) {
    Number *n;

    n = [Number create: [self getZone]];
    [n setUnsigned: 0];
    [landUseCtr addObject: n withKey: [lu atOffset: i]];
  }
}

/*

reportForYear:toFile:

Create a report of the number of parcels that use each land use at the
end of the year. The land use counter associative array must first be
initialised to zero for each land use. Then we loop through the land
parcels and add one to the appropriate land use counter according to
the land use of the parcel. Then we loop through the land uses and print
out the totals.

Note that [FearlusArguments nl] is used as the string to print for a
new line. This is in case DOS mode has been specified on the command
line.

*/

-(void)reportForYear: (unsigned)year toFile: (FILE *)fp {
  id inx, luArray;
  int i;
  LandUse *lu;
  Number *n;
  
  luArray = [[model getEnvironment] getLandUses];
  /* Initialise all the numbers to zero */
  for(i = 0; i < [parameter nUses]; i++) {
    lu = [luArray atOffset: i];
    n = [landUseCtr getObjectWithKey: lu];
    [n setUnsigned: 0];
  }
  /* Count up the land parcels for each sup-population */
  for(inx = [[[model getEnvironment] getLandParcels] begin: [self getZone]],
	[inx next];
      [inx getLoc] == Member;
      [inx next]) {
    lu = [(LandParcel *)[inx get] getLandUse];
    n = [landUseCtr getObjectWithKey: lu];
    [n setUnsigned: [n getUnsigned] + 1];
  }
  [inx drop];
  /* Now loop through the sub-populations and print out the totals */
  for(i = 0; i < [parameter nUses]; i++) {
    lu = [luArray atOffset: i];
    n = [landUseCtr getObjectWithKey: lu];
    fprintf(fp, "Land use ID:\t%u\tLand parcels:\t%u%s", [lu getPIN],
	    [n getUnsigned], [FearlusArguments nl]);
  }
}

/*

setOption:toValue:

setOptionFlag:toValue:

There are no options to this report, so each method will print an error message

*/

-(BOOL)setOption: (char *)option toValue: (char *)value {
  fprintf(stderr, "Invalid format in report configuration file %s\n"
	  "%s: There are no options for this report: (%s=%s).\n",	
	  [FearlusArguments getReporterCFFile], class_get_class_name([self class]),
	  option, value);
  exit(1);
}

-(BOOL)setOptionFlag: (char *)option toValue: (BOOL)value {
  if(value) {
    fprintf(stderr, "Invalid format in report configuration file %s\n"
	    "%s: There are no option flags for this report: (%s).\n",	
	    [FearlusArguments getReporterCFFile], class_get_class_name([self class]),
	    option);
    exit(1);
  }
  else {
    fprintf(stderr, "Invalid format in report configuration file %s\n"
	    "%s: There are no option flags for this report: (No%s).\n",	
	    [FearlusArguments getReporterCFFile], class_get_class_name([self class]),
	    option);
    exit(1);
  }
}

/*

setFixedYears:

Set the fixed years for reporting.

*/

-(void)setFixedYears: (id <List>)y {
  fixedYears = y;
}

/*

setRepeatYears:

Set the repeating years for reporting.

*/

-(void)setRepeatYears: (id <List>)y {
  repeatYears = y;
}

/* There's no need for a drop method -- the Reporter object will own all Report
   objects, and stores them in its own Zone. When the zone is deleted, all the
   Reports' allocated memory will go with it, so long as we use [self getZone]
   here when allocating memory. */

/*

reportingForYear:

Return YES if we're reporting this year, according to the years we've been
handed. NO otherwise.

*/

-(BOOL)reportingForYear: (unsigned)year {
  BOOL reportingYear;
  id inx;
  Number *n;

  /* Decide if we should print this year */
  reportingYear = NO;
  if(([fixedYears getCount] > 0)
     && ([(Number *)[fixedYears atOffset: 0] getUnsigned] == year)) {
    /* This is a fixed year to print. So we'll go ahead, but after we've
       removed the head */
    [[fixedYears removeFirst] drop];
    reportingYear = YES;
  }
  else {
    /* Not a fixed year to print, so check all the repeating years */
    for(inx = [repeatYears begin: scratchZone], [inx next];
	[inx getLoc] == Member;
	[inx next]) {
      n = [inx get];
      if((year % [n getUnsigned]) == 0) {
	/* It's a repeating year to print */
	reportingYear = YES;
	break;
      }
    }
    [inx drop];
  }
  return reportingYear;
}

@end

